/*
 *  input.c
 * Copyright  1992 Niklas Rjemo
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
extern int errno;
#include "error.h"
#include "global.h"
#include "input.h"
#include "riscos.h"
#include "filestack.h"
#include "uname.h"
#include "macros.h"
#include "lex.h"

#define MAX_LINE 4096


extern int gcc_backend;
extern int dde;
extern int objasm;
extern int local;
extern int uc; /* these externs from main.c */

FILE *asmfile;
static char
  buff[MAX_LINE+256], /* 256 bytes of overflow */
  workBuff[MAX_LINE];
static char *pos,*mark;
char *inputName=0;

long int inputLineNo;
BOOL inputExpand=TRUE;
BOOL inputRewind=FALSE;

#if 'A'==65
#define FLIP(x) (isalpha(x) ? (x)^32 : (x))
#define TOLOWER(x) (isupper(x) ? (x)+32 : (x))
  /* these are ASCII-like charset dependent */
#else
#define FLIP(x) (isupper(x) ? tolower(x) : islower(x) ? toupper(x) : (x))
#define TOLOWER(x) tolower(x)
#endif


char inputLook(void)
{ return *pos; }


char inputLookLower(void)
{
  return TOLOWER(*pos);
}

char inputLookUC(void)
{
  char x=*pos;
  return uc ? FLIP(x) : TOLOWER(x);
}


int inputComment(void)
{
  int c=*pos;
  return c==0 || c == ';' || (gcc_backend && c == '@');
}


char inputLookN(int n)  /* Unsafe */
{ return pos[n]; }


char inputLookNLower(int n)  /* Unsafe */
{
  return TOLOWER(pos[n]);
}


char inputLookNUC(int n) /* Unsafe */
{
  char x=pos[n];
  return uc ? FLIP(x) : TOLOWER(x);
}


char inputGet(void)
{ if (*pos) return *pos++; else return *pos; }


char inputGetLower(void)
{
  char c=*pos ? *pos++ : *pos;
  return TOLOWER(c);
}


char inputGetUC(void)
{
  char x=*pos ? *pos++ : *pos;
  return uc ? FLIP(x) : TOLOWER(x);
}


void inputUnGet(char c)
{
  if(pos>buff && pos[-1] == c)
    pos--;
  else
    if (*pos || c) error(ErrorSerious,FALSE,"Internal inputUnGet: illegal character");
}


void inputSkip(void)
{ if (*pos) pos++; }


void inputSkipN(int n)
{
  while (*pos && n--) pos++;
  if (n>(*pos ? -1 : 0)) /* Fix to allow skip to end of string */
    error(ErrorSerious,FALSE,"Internal inputSkipN: trying to skip more characters than are available");
}


char *inputRest(void)
{
  char *t = pos;
  register char *p=pos-1;
  while (*++p); pos=p;
  return t;
}


char *inputLine(void)
{ return buff; }


char inputSkipLook(void)
{ return *pos ? *++pos : *pos; }


void skipblanks(void)
{
  char c;
  register char* p=pos;
  while((c=*p)!=0 && isspace(c)) p++;
  pos=p;
}


void skiprest(void)
{ (pos=mark=buff)[0] = 0; }


void inputMark(void)
{ mark = pos; }


void inputRollback(void)
{ pos = mark; }


void inputInit(char *infile)
{
  static char buffer[16*1024];

  if(infile && strcmp(infile,"-")) {
#ifdef UNIXLIB
    if ((asmfile = fopen (infile, "r")) == NULL)
      errorLine(0,0,ErrorAbort,FALSE,"AS can't read %s: %s",infile, strerror (errno));
#else
    if ((asmfile = fopen (uname (infile, dde), "r")) == NULL)
      errorLine(0,0,ErrorAbort,FALSE,"AS can't read %s",infile);
#endif
    else {
      inputName=CanonicalisePath(infile);
      setvbuf (asmfile, buffer, _IOFBF, 16 * 1024);
    }
  } else
    asmfile = stdin;
/*  macroCurrent=malloc(sizeof(Macro));  sets up a dummy macro
  if (!macroCurrent)
    errorLine(0,0,ErrorAbort,FALSE,"Not enough memory",infile);
  macroCurrent->name="";
  macroCurrent->file=(char*)inputName;
 */
  inputLineNo = 0;
  skiprest();
}


void inputFinish(void)
{ if (asmfile && asmfile!=stdin) fclose(asmfile); }


BOOL inputArgSub(void);


BOOL inputNextLine(void)
{
  int l;
  if (inputRewind) { inputRewind=FALSE; goto ret; }
  if (asmfile==NULL) goto retBad;
  inputLineNo++;
  if (macroCurrent && macroGetLine(workBuff)) goto ret;
  while (fgets(workBuff,MAX_LINE,asmfile)==NULL) {
    if ((asmfile=pop_file())==NULL) {
retBad:
      inputLineNo=-inputLineNo;
      return FALSE;
    }
  }
  l = strlen(workBuff);
  if (l) {
    /* If we start the line with a '#' then treat the whole
     * line as a comment.
     */
    if (workBuff[0] == '#')     {
      (pos=buff)[0]=0;
      return TRUE;
    }
    if(workBuff[l-1] != '\n') {
      if((l=getc(asmfile))!= EOF && l!='\n') {
        error(ErrorSerious,TRUE,"Line truncated");
        while((l=getc(asmfile))!= EOF && l!='\n') ;
      }
    } else
      workBuff[l-1] = 0;
  }
ret:
  if (!inputExpand) {
    strcpy(pos=buff, workBuff);
    return TRUE;
  }
  return inputArgSub();
}


BOOL inputArgSub(void)
{
  Lex label;
  Symbol* sym;
  int ptr=0, trunc=0, len;
  char g, c, *rb;

  g=gcc_backend ? '@' : 0;
  pos=workBuff;
  while (*pos && ptr<MAX_LINE) {
    while (*pos && *pos != '"' && *pos != '\'' && *pos != '|' && *pos != '$'
    	   && *pos != ';' && (*pos != '<' || !objasm) && *pos != g
    	   && ptr < MAX_LINE)
      buff[ptr++] = *pos++;

      switch (c=*pos) {
      case ';': case '@':
        len=strlen(pos);
        if (ptr+len>=MAX_LINE) {
          memcpy(buff+ptr,pos,MAX_LINE-ptr-len);
          goto truncated;
        }
        strcpy(buff+ptr,pos);
        goto finished;
      case '<': /* is it a variable name? */
        rb=pos; while (*++rb!=0 && *rb!='>' && *rb>32);
        if (*rb!='>') { buff[ptr++]=*pos++; break; }
        c='>'; /* set delimiter and fall through */
      case '\'': case '"': case '|':
        do { buff[ptr++]=*pos++; } while (*pos && *pos!=c && ptr<MAX_LINE);
        if (*pos == c) { buff[ptr++] = c; pos++; }
        break;
      case '$':
/*        if (!inputExpand) {
          buff[ptr++]='$';
          if (*++pos=='$') {
            buff[ptr++]='$'; break;
          } else {
            char *c=inputSymbol(&len,0);
            memcpy(buff+ptr,c,len);
            ptr+=len;
          }
          break;
        } else */
        rb=pos; len=ptr;
        if (*++pos=='$') { buff[ptr++]='$'; pos++; break; }
        label=lexGetId();
        if (label.tag!=LexId) {
          error(ErrorWarning,TRUE,"Non-ID in $ expansion");
          buff[ptr++]='$';
          break;
        }
        if (c && *pos=='.') pos++;
        if (local && label.LexId.len==1 && toupper(*label.LexId.str)=='L') {
          buff[ptr++]='$';
          buff[ptr++]=*label.LexId.str;
          break;
        }
        sym=symbolFind(label);
        if (sym) {
          switch (sym->value.Tag.t) {
            case ValueInt:
              sprintf(buff+ptr,"%i",sym->value.ValueInt.i);
              ptr+=strlen(buff+ptr);
              break;
            case ValueFloat:
              sprintf(buff+ptr,"%f",sym->value.ValueFloat.f);
              ptr+=strlen(buff+ptr);
              break;
            case ValueString:
              if (ptr+sym->value.ValueString.len>=MAX_LINE)
                ptr=MAX_LINE+1;
              else {
                strcpy(buff+ptr,sym->value.ValueString.s);
                ptr+=sym->value.ValueString.len;
              }
              break;
            case ValueBool:
              strcpy(buff+ptr,sym->value.ValueBool.b ? "{TRUE}":"{FALSE}");
              ptr+=strlen(buff+ptr);
              break;
            case ValueCode:
            case ValueLateLabel:
            case ValueAddr:
              {
                char c=label.LexId.str[label.LexId.len];
                label.LexId.str[label.LexId.len]=0;
                error(ErrorError,TRUE,"$ expansion '%s' is a pointer",
                      label.LexId.str);
                label.LexId.str[label.LexId.len]=c;
              }
              break;
            default: goto unknown;
          }
          while (*pos && ptr<MAX_LINE) buff[ptr++]=*pos++;
          if (ptr>=MAX_LINE) {
            trunc=1; buff[MAX_LINE-1]=0;
          } else
            buff[ptr]=0;
          ptr=len;
          strcpy(pos=rb,buff+ptr); /* only copy what's necessary */
        } else {
          char c;
unknown:  c=label.LexId.str[label.LexId.len];
          label.LexId.str[label.LexId.len]=0;
          error(ErrorError,TRUE,"Unknown value '%s' for $ expansion",
                label.LexId.str);
          label.LexId.str[label.LexId.len]=c;
          buff[ptr++]='$';
        }
    }
  }
  if (ptr>=MAX_LINE || trunc) {
truncated:
    (pos=buff)[MAX_LINE-1]=0;
    error(ErrorSerious,TRUE,"Line expansion truncated");
    return FALSE;
  }
finished:
  (pos=buff)[ptr]=0;
  return TRUE;
}


char *inputSymbol(int *ilen, char del)
{
  char *res = pos;
  char c;
  register char *p=pos;
  if (del) {
    if ((del=='\'' || del=='"') && objasm) {
      if (del=='\'') ++p; /* some special case stuff... */
      while((c=*p)!=0 && c!=del) p++;
    } else {
      while((c=*p)!=0 && c!=del) {
        p++;
        if (c=='\\' && *p) p++;
      }
    }
  } else {
    while((c=*p)!=0 && (isalnum(c) || c == '_' ||
          (c=='$' && local && (p[1]=='l' || p[1]=='L')
           && p[2]!='.' && p[2]!='_'&& !isalpha(p[2])))) {
      p++;
      if (c=='\\' && *p) p++;
    }
  }
  *ilen = (pos=p)-res;
  return res;
}


void inputThisInstead(char *p)
{
  strcpy(pos=mark=buff, p);
}
